<?php

// Фреймворк ArangoDB
use mirzaev\arangodb\connection,
	mirzaev\arangodb\collection,
	mirzaev\arangodb\document;

// Библиотека для ArangoDB
use ArangoDBClient\Document as _document,
	ArangoDBClient\Cursor,
	ArangoDBClient\Statement as _statement;

// Фреймворк Telegram
use Zanzara\Zanzara;
use Zanzara\Context;
use Zanzara\Config;

// Фреймворк для Google Sheets
use Flow\ETL\Adapter\GoogleSheet\GoogleSheetRange,
	Flow\ETL\Adapter\GoogleSheet\GoogleSheetExtractor,
	Flow\ETL\Adapter\GoogleSheet\Columns,
	Flow\ETL\Flow,
	Flow\ETL\Config as _config,
	Flow\ETL\FlowContext,
	Flow\ETL\Row\Entry,
	Flow\ETL\Row,
	Flow\ETL\DSL\To,
	Flow\ETL\DSL\From;

// Фреймворк для Google API
use Google\Client,
	Google\Service\Sheets,
	Google\Service\Sheets\ValueRange;


require __DIR__ . '/../../../../../../../vendor/autoload.php';

$arangodb = new connection(require __DIR__ . '/../settings/arangodb.php');

/* ini_set('error_reporting', E_ALL);
ini_set('display_errors', 1);
ini_set('display_startup_errors', 1); */

$config = new Config();
$config->setParseMode(Config::PARSE_MODE_MARKDOWN);

$bot = new Zanzara(require(__DIR__ . '/../settings/key.php'), $config);

$bot->onMessage(function (Context $ctx) use ($arangodb) {
	// Инициализация сообщения
	$message = $ctx->getMessage();

	if ($members = $ctx->getMessage()->getNewChatMembers()) {
		// Новый аккаунт в чате

		// Инициализация идентификатора сотрудника
		$id = $members[0]->getId();
	} else if ($ctx->getMessage()->getLeftChatMember()) return;
	else {
		// Любое сообщение

		// Инициализация идентификатора отправителя сообщения
		$id = $message->getFrom()->getId();
	}

	if (collection::init($arangodb->session, 'telegram')) {
		// Инициализирована коллекция telegram-аккаунтов в ArangoDB

		if (!$telegram = collection::search($arangodb->session, sprintf("FOR d IN telegram FILTER d.id == %s && !d.deleted RETURN d", $id))) {
			// Не найден telegram-аккаунт в базе данных

			// Запись telegram-аккаунта в базу данных (не привязан сотрудник и активность - прямо сейчас)
			document::write($arangodb->session,	'telegram', ['id' => $id, 'worker' => null, 'activity' => time(), 'welcome' => false]);

			if (!$telegram = collection::search($arangodb->session, sprintf("FOR d IN telegram FILTER d.id == %s && !d.deleted RETURN d", $id))) {
				// Не найден telegram-аккаунт в базе данных

				throw new exception('чё бля происходит');
			}
		}

		if ($telegram->worker === null) {
			// Не авторизован telegram-аккаунт

			if ($telegram->welcome && $telegram->activity - time() < 0 && !$telegram->ban) {
				// Повторный запрос авторизации + прошло более 10800 секунд (3 часа) с последней активности (подразумевается отправка сообщения)
				// Подразумевается, что если $telegram->ban === true, то уже идёт процесс авторизации и это сообщение уже было отправлено

				$ctx->sendMessage("⚠️ Введите ваш табельный номер\nТребуется дополнительная проверка", ['reply_to_message_id' => $ctx->getMessage()->getMessageId()])
					->then(function ($message) use ($ctx, $id, $arangodb, $telegram) {
						// Запись в реестр сообщений
						$telegram->messages = [
							$message->getChat()->getId() => [
								$message->getMessageId() =>	$message->getText()
							] + ($telegram->messages[$message->getChat()->getId()] ?? [])
						] + ($telegram->messages ?? []);

						// Инициализация таймера изгнания из чата (180 секунд - 3 минуты)
						$telegram->ban = time() + 180;

						// Запись в ArangoDB
						document::update($arangodb->session, $telegram);

						// Инициализация таймера изгнания сотрудника и удаления сообщения
						React\Promise\Timer\sleep(180)->then(
							function () use ($ctx, $id, $message, $arangodb) {
								if ($telegram = collection::search($arangodb->session, sprintf("FOR d IN telegram FILTER d.id == %s RETURN d", $id))) {
									// Найден telegram-аккаунт в базе данных (повторное чтение для обновления буфера сообщений)

									if ($telegram->messages[(string) $message->getChat()->getId()][(string) $message->getMessageId()]) {
										// Найдено сообщение в ArangoDB (подразумевается, что оно есть и в чате)

										// Удаление сообщения из чата
										$ctx->deleteMessage($message->getChat()->getId(), $message->getMessageId());

										// Удаление сообщения из ArangoDB
										$buffer = $telegram->messages;
										unset($buffer[(string) $message->getChat()->getId()][(string) $message->getMessageId()]);

										// По первому уровню массива (чату) ArangoDB думает, что изменений нет и не перезаписывает (лень разбираться)
										$telegram->messages = [];
										document::update($arangodb->session, $telegram);

										$telegram->messages = $buffer;

										// Запись в ArangoDB
										document::update($arangodb->session, $telegram);
									}

									if ($telegram->ban) {
										// Прошло более 180 секунд (3 минуты) без авторизации после запроса

										// Изгнание из чата
										$ctx->kickChatMember($message->getChat()->getId(), $id)
											->then(function () use ($ctx, $id, $message, $arangodb, $telegram) {
												// Изгнан из чата

												// Удаление всех сообщений из чата
												foreach ($telegram->messages ?? [] as $chat => $messages)
													foreach ($messages as $_message => $text)
														$ctx->deleteMessage($chat, $_message);

												// Удаление всех сообщений в чате из ArangoDB
												$telegram->messages = [(string) $message->getChat()->getId() => []] + ($telegram->messages ?? []);

												// Запись статуса удаления в ArangoDB
												$telegram->deleted = true;

												// Запись таймера изгнания
												$telegram->ban = false;

												// Запись в ArangoDB
												document::update($arangodb->session, $telegram);
											});
									}
								}
							},
							fn () => null
						);
					});

				// Возврат (успех)
				return;
			} else if (!$telegram->welcome) {
				// Первый запрос авторизации

				$ctx->sendMessage("⚠️ Введите ваш табельный номер", ['reply_to_message_id' => $ctx->getMessage()->getMessageId()])
					->then(function ($message) use ($ctx, $id, $arangodb, $telegram) {
						// Запись статуса первого запроса авторизации
						$telegram->welcome = true;

						// Запись в реестр сообщений
						$telegram->messages = [
							$message->getChat()->getId() => [
								$message->getMessageId() =>	$message->getText()
							] + ($telegram->messages[$message->getChat()->getId()] ?? [])
						] + ($telegram->messages ?? []);

						// Инициализация таймера изгнания из чата (180 секунд - 3 минуты)
						$telegram->ban = time() + 180;

						// Запись в ArangoDB
						document::update($arangodb->session, $telegram);

						// Инициализация таймера изгнания сотрудника и удаления сообщения
						React\Promise\Timer\sleep(180)->then(
							function () use ($ctx, $id, $message, $arangodb) {
								if ($telegram = collection::search($arangodb->session, sprintf("FOR d IN telegram FILTER d.id == %s RETURN d", $id))) {
									// Найден telegram-аккаунт в базе данных (повторное чтение для обновления буфера сообщений)

									if ($telegram->messages[(string) $message->getChat()->getId()][(string) $message->getMessageId()]) {
										// Найдено сообщение в ArangoDB (подразумевается, что оно есть и в чате)

										// Удаление сообщения из чата
										$ctx->deleteMessage($message->getChat()->getId(), $message->getMessageId());

										// Удаление сообщения из ArangoDB
										$buffer = $telegram->messages;
										unset($buffer[(string) $message->getChat()->getId()][(string) $message->getMessageId()]);

										// По первому уровню массива (чату) ArangoDB думает, что изменений нет и не перезаписывает (лень разбираться)
										$telegram->messages = [];
										document::update($arangodb->session, $telegram);

										$telegram->messages = $buffer;

										// Запись в ArangoDB
										document::update($arangodb->session, $telegram);
									}

									if ($telegram->ban) {
										// Прошло более 180 секунд (3 минуты) без авторизации после завпроса

										// Изгнание из чата
										$ctx->kickChatMember($message->getChat()->getId(), $id)
											->then(function () use ($ctx, $id, $message, $arangodb, $telegram) {
												// Изгнан из чата

												// Удаление всех сообщений из чата
												foreach ($telegram->messages ?? [] as $chat => $messages)
													foreach ($messages as $_message => $text)
														$ctx->deleteMessage($chat, $_message);

												// Удаление всех сообщений в чате из ArangoDB
												$telegram->messages = [(string) $message->getChat()->getId() => []] + ($telegram->messages ?? []);

												// Запись статуса удаления в ArangoDB
												$telegram->deleted = true;

												// Запись таймера изгнания
												$telegram->ban = false;

												// Запись в ArangoDB
												document::update($arangodb->session, $telegram);
											});
									}
								}
							},
							fn () => null
						);
					});

				// Возврат (успех)
				return;
			}

			if ($telegram->ban && $telegram->ban - time() < 0) {
				// Прошло более 180 секунд (3 минуты) без авторизации после завпроса

				// Изгнание из чата
				$ctx->kickChatMember($message->getChat()->getId(), $id)
					->then(function () use ($ctx, $id, $message, $arangodb, $telegram) {
						// Изгнан из чата

						// Удаление всех сообщений из чата
						foreach ($telegram->messages ?? [] as $chat => $messages)
							foreach ($messages as $_message => $text)
								$ctx->deleteMessage($chat, $_message);

						// Удаление всех сообщений в чате из ArangoDB
						$telegram->messages = [(string) $message->getChat()->getId() => []] + ($telegram->messages ?? []);

						// Запись статуса удаления в ArangoDB
						$telegram->deleted = true;

						// Запись таймера изгнания
						$telegram->ban = false;

						// Запись в ArangoDB
						document::update($arangodb->session, $telegram);
					});

				// Возврат (успех)
				return;
			} else {
				// Не прошло 180 секунд (3 минуты) без авторизации после запроса (либо запроса ещё не было)

				// Инициализация переданного идентификатора (только цифры)
				$worker = preg_replace('/[^\d]/', '', $message->getText() ?? '');

				if (!empty($worker)) {
					// Инициализирован идентификатор сотрудника для поиска

					$settings = json_decode(require(__DIR__ . '/../settings/workers/google.php'), true);
					$document = require(__DIR__ . '/../settings/workers/document.php');
					$sheets = require(__DIR__ . '/../settings/workers/sheets.php');

					$client = new Client();
					$client->setScopes(Sheets::SPREADSHEETS);
					$client->setAuthConfig($settings);
					$api = new Sheets($client);

					foreach ($sheets as $sheet) {
						$rows = (new Flow())->read(new GoogleSheetExtractor($api, $document, new Columns($sheet, 'A', 'A'), true, 1000));

						foreach ($rows->fetch(10000) as $row) {
							// Перебор строк

							if ($worker === $row->toArray()['ID']) {
								// Найден идентификатор

								if (collection::search($arangodb->session, sprintf("FOR d IN telegram FILTER d.worker == %s RETURN d", $worker)))
									$ctx->sendMessage("🪪 Сотрудник \\($worker\\) уже привязан к другому telegram\\-аккаунту", ['reply_to_message_id' => $message->getMessageId()])
										->then(
											function ($message) use ($ctx, $id, $arangodb, $telegram) {
												// Запись в реестр сообщений
												$telegram->messages = [
													$message->getChat()->getId() => [
														$message->getMessageId() =>	$message->getText()
													] + ($telegram->messages[$message->getChat()->getId()] ?? [])
												] + ($telegram->messages ?? []);

												// Запись в ArangoDB
												document::update($arangodb->session, $telegram);

												// Инициализация таймера удаления сообщения
												React\Promise\Timer\sleep(10)->then(
													function () use ($ctx, $id, $message, $arangodb) {
														if ($telegram = collection::search($arangodb->session, sprintf("FOR d IN telegram FILTER d.id == %s RETURN d", $id))) {
															// Найден telegram-аккаунт в базе данных (повторное чтение для обновления буфера сообщений)

															if ($telegram->messages[(string) $message->getChat()->getId()][(string) $message->getMessageId()]) {
																// Найдено сообщение в ArangoDB (подразумевается, что оно есть и в чате)

																// Удаление сообщения из чата
																$ctx->deleteMessage($message->getChat()->getId(), $message->getMessageId());

																// Удаление сообщения из ArangoDB
																$buffer = $telegram->messages;
																unset($buffer[(string) $message->getChat()->getId()][(string) $message->getMessageId()]);

																// По первому уровню массива (чату) ArangoDB думает, что изменений нет и не перезаписывает (лень разбираться)
																$telegram->messages = [];
																document::update($arangodb->session, $telegram);

																$telegram->messages = $buffer;

																// Запись в ArangoDB
																document::update($arangodb->session, $telegram);
															}
														}
													},
													fn () => null
												);
											}
										);
								else {
									// Не занят идентификатор

									// Запись статуса изгнания (остановка процесса)
									$telegram->ban = false;

									// Запись сотрудника (привязка)
									$telegram->worker = (int) $worker;

									// Удаление всех сообщений из чата
									foreach ($telegram->messages ?? [] as $chat => $messages)
										foreach ($messages as $_message => $text)
											$ctx->deleteMessage($chat, $_message);

									// Удаление всех сообщений в чате из ArangoDB
									$telegram->messages = [(string) $message->getChat()->getId() => []] + ($telegram->messages ?? []);

									// Запись в ArangoDB
									document::update($arangodb->session, $telegram);

									$ctx->sendMessage("✅ Авторизован сотрудник: $worker", ['reply_to_message_id' => $message->getMessageId()])
										->then(
											function ($message) use ($ctx, $id, $arangodb, $telegram) {
												// Запись в реестр сообщений
												$telegram->messages = [
													$message->getChat()->getId() => [
														$message->getMessageId() =>	$message->getText()
													] + ($telegram->messages[$message->getChat()->getId()] ?? [])
												] + ($telegram->messages ?? []);

												// Запись в ArangoDB
												document::update($arangodb->session, $telegram);

												// Инициализация таймера удаления сообщения
												React\Promise\Timer\sleep(10)->then(
													function () use ($ctx, $id, $message, $arangodb) {
														if ($telegram = collection::search($arangodb->session, sprintf("FOR d IN telegram FILTER d.id == %s RETURN d", $id))) {
															// Найден telegram-аккаунт в базе данных (повторное чтение для обновления буфера сообщений)

															if ($telegram->messages[(string) $message->getChat()->getId()][(string) $message->getMessageId()]) {
																// Найдено сообщение в ArangoDB (подразумевается, что оно есть и в чате)

																// Удаление сообщения из чата
																$ctx->deleteMessage($message->getChat()->getId(), $message->getMessageId());

																// Удаление сообщения из ArangoDB
																$buffer = $telegram->messages;
																unset($buffer[(string) $message->getChat()->getId()][(string) $message->getMessageId()]);

																// По первому уровню массива (чату) ArangoDB думает, что изменений нет и не перезаписывает (лень разбираться)
																$telegram->messages = [];
																document::update($arangodb->session, $telegram);

																$telegram->messages = $buffer;

																// Запись в ArangoDB
																document::update($arangodb->session, $telegram);
															}
														}
													},
													fn () => null
												);
											}
										);
								}

								// Возврат (успех)
								return;
							}
						}
					}

					$ctx->sendMessage("⛔ Не найден сотрудник: $worker", ['reply_to_message_id' => $message->getMessageId()])
						->then(
							function ($message) use ($ctx, $id, $arangodb, $telegram) {
								// Запись в реестр сообщений
								$telegram->messages = [
									$message->getChat()->getId() => [
										$message->getMessageId() =>	$message->getText()
									] + ($telegram->messages[$message->getChat()->getId()] ?? [])
								] + ($telegram->messages ?? []);

								// Запись в ArangoDB
								document::update($arangodb->session, $telegram);

								// Инициализация таймера
								React\Promise\Timer\sleep(10)->then(
									function () use ($ctx, $id, $message, $arangodb) {
										if ($telegram = collection::search($arangodb->session, sprintf("FOR d IN telegram FILTER d.id == %s RETURN d", $id))) {
											// Найден telegram-аккаунт в базе данных (повторное чтение для обновления буфера сообщений)

											if ($telegram->messages[(string) $message->getChat()->getId()][(string) $message->getMessageId()]) {
												// Найдено сообщение в ArangoDB (подразумевается, что оно есть и в чате)

												// Удаление сообщения из чата
												$ctx->deleteMessage($message->getChat()->getId(), $message->getMessageId());

												// Удаление сообщения из ArangoDB
												$buffer = $telegram->messages;
												unset($buffer[(string) $message->getChat()->getId()][(string) $message->getMessageId()]);

												// По первому уровню массива (чату) ArangoDB думает, что изменений нет и не перезаписывает (лень разбираться)
												$telegram->messages = [];
												document::update($arangodb->session, $telegram);

												$telegram->messages = $buffer;

												// Запись в ArangoDB
												document::update($arangodb->session, $telegram);
											}
										}
									},
									fn () => null
								);
							}
						);

					// Возврат (успех)
					return;
				}
			}
		}

		// Запись активности
		$telegram->activity = time();
		document::update($arangodb->session, $telegram);
	}
});

$bot->run();
